#!/usr/bin/python3
#
# File:    ./setup.py
# Author:  Jiří Kučera <sanczes AT gmail.com>
# Date:    2022-06-23 10:01:39 +0200
# Project: vutils-yaml: Working with YAML format
#
# SPDX-License-Identifier: MIT
#
"""Setup for vutils-yaml."""

from setuptools import setup

setup()
